// <author>Netmon Cipher Dev Team</author>
// <date>2009-07-16</date>
// <version>1.0</version>
// 
// <summary>
// This contains methods to Decrypt the Encrypted AES Application Data
// transferred over the SSL layer.
// </summary>

namespace AesLib
{
    using System;

    /// <summary>
    /// Advanced Encryption Standard Class
    /// </summary>
    public class AesClass
    {
        /// <summary>
        /// Block size in 32-bit words.  Always 4 for AES.  (128 bits).
        /// </summary>
        private int nb;

        /// <summary>
        /// Key size in 32-bit words.  4, 6, 8.  (128, 192, 256 bits).
        /// </summary>
        private int nk;

        /// <summary>
        /// Number of rounds. 10, 12, 14.
        /// </summary> 
        private int nr;

        /// <summary>
        /// Seed key. size will be 4 * keySize from ctor.
        /// </summary>
        private byte[] key;

        /// <summary>
        /// Substitution box
        /// </summary>
        private byte[,] sbox;

        /// <summary>
        /// Inverse Substitution box
        /// </summary>
        private byte[,] invSbox;

        /// <summary>
        /// Key schedule array.
        /// </summary>
        private byte[,] w;

        /// <summary>
        /// Round constants.
        /// </summary>
        private byte[,] rcon;

        /// <summary>
        /// State matrix
        /// </summary> 
        private byte[,] state;

        /// <summary>
        /// Initializes a new instance of the AesClass class.
        /// </summary>
        /// <param name="keySize">Key Size for Aes</param>
        /// <param name="keyBytes">Key Byte Array</param>
        public AesClass(KeySize keySize, byte[] keyBytes)
        {
            this.SetNbNkNr(keySize);

            this.key = new byte[this.nk * 4];  //// 16, 24, 32 bytes
            keyBytes.CopyTo(this.key, 0);

            this.BuildSbox();
            this.BuildInvSbox();
            this.BuildRcon();

            //// expand the seed key into a key schedule and store in w
            this.KeyExpansion();
        }

        /// <summary>
        /// Key size, in bits, for constructor
        /// </summary>
        public enum KeySize
        {
            /// <summary>
            /// 128 Bit AES
            /// </summary>
            Bits128,

            /// <summary>
            /// 192 Bit AES
            /// </summary>
            Bits192,

            /// <summary>
            /// 256 Bit AES
            /// </summary>
            Bits256
        }

        /// <summary>
        /// Encipher 16-bit input
        /// </summary>
        /// <param name="input">input byte array</param>
        /// <param name="output">output byte array</param>
        public void Cipher(byte[] input, byte[] output)
        {
            //// state = input
            this.state = new byte[4, this.nb];  //// always [4,4]
            for (int i = 0; i < (4 * this.nb); ++i)
            {
                this.state[i % 4, i / 4] = input[i];
            }

            this.AddRoundKey(0);

            //// main round loop
            for (int round = 1; round <= (this.nr - 1); ++round)
            {
                this.SubBytes();
                this.ShiftRows();
                this.MixColumns();
                this.AddRoundKey(round);
            }

            this.SubBytes();
            this.ShiftRows();
            this.AddRoundKey(this.nr);

            //// output = state
            for (int i = 0; i < (4 * this.nb); ++i)
            {
                output[i] = this.state[i % 4, i / 4];
            }
        }

        /// <summary>
        /// Decipher 16-bit input
        /// </summary>
        /// <param name="input">input byte array</param>
        /// <param name="output">output byte array</param>
        public void InvCipher(byte[] input, byte[] output)
        {
            //// state = input
            this.state = new byte[4, this.nb];  //// always [4,4]
            for (int i = 0; i < (4 * this.nb); ++i)
            {
                this.state[i % 4, i / 4] = input[i];
            }

            this.AddRoundKey(this.nr);

            //// main round loop
            for (int round = this.nr - 1; round >= 1; --round)
            {
                this.InvShiftRows();
                this.InvSubBytes();
                this.AddRoundKey(round);
                this.InvMixColumns();
            }

            this.InvShiftRows();
            this.InvSubBytes();
            this.AddRoundKey(0);

            //// output = state
            for (int i = 0; i < (4 * this.nb); ++i)
            {
                output[i] = this.state[i % 4, i / 4];
            }
        }

        /// <summary>
        /// Prints on Console.
        /// </summary>
        public void Dump()
        {
            ////Console.WriteLine("Nb = " + Nb + " Nk = " + Nk + " Nr = " + Nr);
            ////Console.WriteLine("\nThe key is \n" + DumpKey() );
            ////Console.WriteLine("\nThe Sbox is \n" + DumpTwoByTwo(Sbox));
            ////Console.WriteLine("\nThe w array is \n" + DumpTwoByTwo(w));
            ////Console.WriteLine("\nThe State array is \n" + DumpTwoByTwo(State));
        }

        /// <summary>
        /// Evaluates Dump Kay
        /// </summary>
        /// <returns>Returns the key string</returns>
        public string DumpKey()
        {
            string s = string.Empty;
            for (int i = 0; i < this.key.Length; ++i)
            {
                s += this.key[i].ToString("x2") + " ";
            }

            return s;
        }

        /// <summary>
        /// Dump Two By Two
        /// </summary>
        /// <param name="a">2D Array to dump</param>
        /// <returns>Returns Key String</returns>
        public string DumpTwoByTwo(byte[,] a)
        {
            string s = string.Empty;
            for (int r = 0; r < a.GetLength(0); ++r)
            {
                s += "[" + r + "]" + " ";
                for (int c = 0; c < a.GetLength(1); ++c)
                {
                    s += a[r, c].ToString("x2") + " ";
                }

                s += "\n";
            }

            return s;
        }

        /// <summary>
        /// gf multiplication by 1
        /// </summary>
        /// <param name="b">Byte to multiply</param>
        /// <returns>Returns the resultant byte</returns>
        private static byte GFmultby01(byte b)
        {
            return b;
        }

        /// <summary>
        /// gf multiplication by 2
        /// </summary>
        /// <param name="b">Byte to multiply</param>
        /// <returns>Returns the resultant byte</returns>
        private static byte GFmultby02(byte b)
        {
            if (b < 0x80)
            {
                return (byte)(int)(b << 1);
            }
            else
            {
                return (byte)((int)(b << 1) ^ (int)0x1b);
            }
        }

        /// <summary>
        /// gf multiplication by 3
        /// </summary>
        /// <param name="b">Byte to multiply</param>
        /// <returns>Returns the resultant byte</returns>
        private static byte GFmultby03(byte b)
        {
            return (byte)((int)GFmultby02(b) ^ (int)b);
        }

        /// <summary>
        /// gf multiplication by 9
        /// </summary>
        /// <param name="b">Byte to multiply</param>
        /// <returns>Returns the resultant byte</returns>
        private static byte GFmultby09(byte b)
        {
            return (byte)((int)GFmultby02(GFmultby02(GFmultby02(b))) ^
                           (int)b);
        }

        /// <summary>
        /// gf multiplication by 0x0b
        /// </summary>
        /// <param name="b">Byte to multiply</param>
        /// <returns>Returns the resultant byte</returns>
        private static byte GFmultby0b(byte b)
        {
            return (byte)((int)GFmultby02(GFmultby02(GFmultby02(b))) ^
                           (int)GFmultby02(b) ^
                           (int)b);
        }

        /// <summary>
        /// gf multiplication by 0x0d
        /// </summary>
        /// <param name="b">Byte to multiply</param>
        /// <returns>Returns the resultant byte</returns>
        private static byte GFmultby0d(byte b)
        {
            return (byte)((int)GFmultby02(GFmultby02(GFmultby02(b))) ^
                           (int)GFmultby02(GFmultby02(b)) ^
                           (int)b);
        }

        /// <summary>
        /// gf multiplication by 0x0e
        /// </summary>
        /// <param name="b">Byte to multiply</param>
        /// <returns>Returns the resultant byte</returns>
        private static byte GFmultby0e(byte b)
        {
            return (byte)((int)GFmultby02(GFmultby02(GFmultby02(b))) ^
                           (int)GFmultby02(GFmultby02(b)) ^
                           (int)GFmultby02(b));
        }

        /// <summary>
        /// Sets values for Nb, Nk and Nr
        /// </summary>
        /// <param name="keySize">Key Size Value</param>
        private void SetNbNkNr(KeySize keySize)
        {
            this.nb = 4;     //// block size always = 4 words = 16 bytes = 128 bits for AES

            if (keySize == KeySize.Bits128)
            {
                this.nk = 4;   //// key size = 4 words = 16 bytes = 128 bits
                this.nr = 10;  //// rounds for algorithm = 10
            }
            else if (keySize == KeySize.Bits192)
            {
                this.nk = 6;   //// 6 words = 24 bytes = 192 bits
                this.nr = 12;
            }
            else if (keySize == KeySize.Bits256)
            {
                this.nk = 8;   //// 8 words = 32 bytes = 256 bits
                this.nr = 14;
            }
        }

        /// <summary>
        /// Builds SBox for AES
        /// </summary>
        private void BuildSbox()
        {
            //// populate the Sbox matrix
            this.sbox = new byte[16, 16] 
            {
                /* 0     1     2     3     4     5     6     7     8     9     a     b     c     d     e     f */
        /*0*/ { 0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76 },
        /*1*/ { 0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0 },
        /*2*/ { 0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15 },
        /*3*/ { 0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75 },
        /*4*/ { 0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84 },
        /*5*/ { 0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf },
        /*6*/ { 0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8 },
        /*7*/ { 0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2 },
        /*8*/ { 0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73 },
        /*9*/ { 0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb },
        /*a*/ { 0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79 },
        /*b*/ { 0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08 },
        /*c*/ { 0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a },
        /*d*/ { 0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e },
        /*e*/ { 0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf },
        /*f*/ { 0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16 }
            };
        }

        /// <summary>
        /// Builds Inverse SBox for AES
        /// </summary>
        private void BuildInvSbox()
        {
            //// populate the iSbox matrix
            this.invSbox = new byte[16, 16] 
            {
                /* 0     1     2     3     4     5     6     7     8     9     a     b     c     d     e     f */
        /*0*/ { 0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e, 0x81, 0xf3, 0xd7, 0xfb },
        /*1*/ { 0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87, 0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb },
        /*2*/ { 0x54, 0x7b, 0x94, 0x32, 0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e },
        /*3*/ { 0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49, 0x6d, 0x8b, 0xd1, 0x25 },
        /*4*/ { 0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92 },
        /*5*/ { 0x6c, 0x70, 0x48, 0x50, 0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84 },
        /*6*/ { 0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05, 0xb8, 0xb3, 0x45, 0x06 },
        /*7*/ { 0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02, 0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b },
        /*8*/ { 0x3a, 0x91, 0x11, 0x41, 0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73 },
        /*9*/ { 0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8, 0x1c, 0x75, 0xdf, 0x6e },
        /*a*/ { 0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89, 0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b },
        /*b*/ { 0xfc, 0x56, 0x3e, 0x4b, 0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4 },
        /*c*/ { 0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xec, 0x5f },
        /*d*/ { 0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d, 0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef },
        /*e*/ { 0xa0, 0xe0, 0x3b, 0x4d, 0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61 },
        /*f*/ { 0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d }
            };
        }

        /// <summary>
        /// Builds Rcon Array
        /// </summary>
        private void BuildRcon()
        {
            this.rcon = new byte[11, 4] 
                {
                    { 0x00, 0x00, 0x00, 0x00 },  
                    { 0x01, 0x00, 0x00, 0x00 },
                    { 0x02, 0x00, 0x00, 0x00 },
                    { 0x04, 0x00, 0x00, 0x00 },
                    { 0x08, 0x00, 0x00, 0x00 },
                    { 0x10, 0x00, 0x00, 0x00 },
                    { 0x20, 0x00, 0x00, 0x00 },
                    { 0x40, 0x00, 0x00, 0x00 },
                    { 0x80, 0x00, 0x00, 0x00 },
                    { 0x1b, 0x00, 0x00, 0x00 },
                    { 0x36, 0x00, 0x00, 0x00 }
                };
        }

        /// <summary>
        /// Adds the round key after every round
        /// </summary>
        /// <param name="round">Indicates the Round Count</param>
        private void AddRoundKey(int round)
        {
            for (int r = 0; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    this.state[r, c] = (byte)((int)this.state[r, c] ^ (int)this.w[(round * 4) + c, r]);
                }
            }
        }

        /// <summary>
        /// Calculates the SubBytes
        /// </summary>
        private void SubBytes()
        {
            for (int r = 0; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    this.state[r, c] = this.sbox[(this.state[r, c] >> 4), (this.state[r, c] & 0x0f)];
                }
            }
        }

        /// <summary>
        /// Calculates the Inverse SubBytes
        /// </summary>
        private void InvSubBytes()
        {
            for (int r = 0; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    this.state[r, c] = this.invSbox[(this.state[r, c] >> 4), (this.state[r, c] & 0x0f)];
                }
            }
        }

        /// <summary>
        /// Shifts the Row
        /// </summary>
        private void ShiftRows()
        {
            byte[,] temp = new byte[4, 4];

            //// copy State into temp[]
            for (int r = 0; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    temp[r, c] = this.state[r, c];
                }
            }

            //// shift temp into State
            for (int r = 1; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    this.state[r, c] = temp[r, (c + r) % this.nb];
                }
            }
        }

        /// <summary>
        /// Shifts the row for inverse
        /// </summary>
        private void InvShiftRows()
        {
            byte[,] temp = new byte[4, 4];

            //// copy State into temp[]
            for (int r = 0; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    temp[r, c] = this.state[r, c];
                }
            }

            //// shift temp into State
            for (int r = 1; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    this.state[r, (c + r) % this.nb] = temp[r, c];
                }
            }
        }

        /// <summary>
        /// Mix Columns
        /// </summary>
        private void MixColumns()
        {
            byte[,] temp = new byte[4, 4];

            //// copy State into temp[]
            for (int r = 0; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    temp[r, c] = this.state[r, c];
                }
            }

            for (int c = 0; c < 4; ++c)
            {
                this.state[0, c] = (byte)((int)GFmultby02(temp[0, c]) ^ (int)GFmultby03(temp[1, c]) ^
                                           (int)GFmultby01(temp[2, c]) ^ (int)GFmultby01(temp[3, c]));
                this.state[1, c] = (byte)((int)GFmultby01(temp[0, c]) ^ (int)GFmultby02(temp[1, c]) ^
                                           (int)GFmultby03(temp[2, c]) ^ (int)GFmultby01(temp[3, c]));
                this.state[2, c] = (byte)((int)GFmultby01(temp[0, c]) ^ (int)GFmultby01(temp[1, c]) ^
                                           (int)GFmultby02(temp[2, c]) ^ (int)GFmultby03(temp[3, c]));
                this.state[3, c] = (byte)((int)GFmultby03(temp[0, c]) ^ (int)GFmultby01(temp[1, c]) ^
                                           (int)GFmultby01(temp[2, c]) ^ (int)GFmultby02(temp[3, c]));
            }
        }

        /// <summary>
        /// Inverse Mix Columns
        /// </summary>
        private void InvMixColumns()
        {
            byte[,] temp = new byte[4, 4];

            //// copy State into temp[]
            for (int r = 0; r < 4; ++r)
            {
                for (int c = 0; c < 4; ++c)
                {
                    temp[r, c] = this.state[r, c];
                }
            }

            for (int c = 0; c < 4; ++c)
            {
                this.state[0, c] = (byte)((int)GFmultby0e(temp[0, c]) ^ (int)GFmultby0b(temp[1, c]) ^
                                           (int)GFmultby0d(temp[2, c]) ^ (int)GFmultby09(temp[3, c]));
                this.state[1, c] = (byte)((int)GFmultby09(temp[0, c]) ^ (int)GFmultby0e(temp[1, c]) ^
                                           (int)GFmultby0b(temp[2, c]) ^ (int)GFmultby0d(temp[3, c]));
                this.state[2, c] = (byte)((int)GFmultby0d(temp[0, c]) ^ (int)GFmultby09(temp[1, c]) ^
                                           (int)GFmultby0e(temp[2, c]) ^ (int)GFmultby0b(temp[3, c]));
                this.state[3, c] = (byte)((int)GFmultby0b(temp[0, c]) ^ (int)GFmultby0d(temp[1, c]) ^
                                           (int)GFmultby09(temp[2, c]) ^ (int)GFmultby0e(temp[3, c]));
            }
        }

        /// <summary>
        /// Key Expansion Method
        /// </summary>
        private void KeyExpansion()
        {
            //// 4 columns of bytes corresponds to a word
            this.w = new byte[this.nb * (this.nr + 1), 4];

            for (int row = 0; row < this.nk; ++row)
            {
                this.w[row, 0] = this.key[4 * row];
                this.w[row, 1] = this.key[(4 * row) + 1];
                this.w[row, 2] = this.key[(4 * row) + 2];
                this.w[row, 3] = this.key[(4 * row) + 3];
            }

            byte[] temp = new byte[4];

            for (int row = this.nk; row < this.nb * (this.nr + 1); ++row)
            {
                temp[0] = this.w[row - 1, 0];
                temp[1] = this.w[row - 1, 1];
                temp[2] = this.w[row - 1, 2];
                temp[3] = this.w[row - 1, 3];

                if (row % this.nk == 0)
                {
                    temp = this.SubWord(this.RotWord(temp));

                    temp[0] = (byte)((int)temp[0] ^ (int)this.rcon[row / this.nk, 0]);
                    temp[1] = (byte)((int)temp[1] ^ (int)this.rcon[row / this.nk, 1]);
                    temp[2] = (byte)((int)temp[2] ^ (int)this.rcon[row / this.nk, 2]);
                    temp[3] = (byte)((int)temp[3] ^ (int)this.rcon[row / this.nk, 3]);
                }
                else if (this.nk > 6 && (row % this.nk == 4))
                {
                    temp = this.SubWord(temp);
                }

                //// w[row] = w[row-Nk] xor temp
                this.w[row, 0] = (byte)((int)this.w[row - this.nk, 0] ^ (int)temp[0]);
                this.w[row, 1] = (byte)((int)this.w[row - this.nk, 1] ^ (int)temp[1]);
                this.w[row, 2] = (byte)((int)this.w[row - this.nk, 2] ^ (int)temp[2]);
                this.w[row, 3] = (byte)((int)this.w[row - this.nk, 3] ^ (int)temp[3]);
            }
        }

        /// <summary>
        /// Evaluates Sub Word
        /// </summary>
        /// <param name="word">Word to Evaluate</param>
        /// <returns>Returns the evaluated sub word</returns>
        private byte[] SubWord(byte[] word)
        {
            byte[] result = new byte[4];
            result[0] = this.sbox[word[0] >> 4, word[0] & 0x0f];
            result[1] = this.sbox[word[1] >> 4, word[1] & 0x0f];
            result[2] = this.sbox[word[2] >> 4, word[2] & 0x0f];
            result[3] = this.sbox[word[3] >> 4, word[3] & 0x0f];
            return result;
        }

        /// <summary>
        /// Evaluates Rot Word
        /// </summary>
        /// <param name="word">Word to Evaluate</param>
        /// <returns>Returns the evaluated rot word</returns>
        private byte[] RotWord(byte[] word)
        {
            byte[] result = new byte[4];
            result[0] = word[1];
            result[1] = word[2];
            result[2] = word[3];
            result[3] = word[0];
            return result;
        }        
    }
}
